# -*- coding: utf-8 -*-
# 
# Copyright (c) 2025~2999 - Cologler <skyoflw@gmail.com>
# ----------
# 
# ----------

from contextlib import nullcontext
from threading import RLock
from typing import Any, Hashable, Iterable, final, override

from .._bases import Factory, IServiceInfo, IServiceProvider, LifeTime
from ..symbols import Symbols

_NULL_CONTEXT = nullcontext()


@final
class ProviderServiceInfo(IServiceInfo[IServiceProvider]):
    '''
    Get current `ServiceProvider`.
    '''
    _INSTANCE: 'ProviderServiceInfo'
    __slots__ = ()

    def __repr__(self) -> str:
        return '<(ioc) => ioc>'

    @override
    def get_service(self, provider: IServiceProvider) -> IServiceProvider:
        return provider

    @classmethod
    def get_singleton_instance(cls) -> 'ProviderServiceInfo':
        return cls._INSTANCE

ProviderServiceInfo._INSTANCE = ProviderServiceInfo()


class GetAttrServiceInfo[T, TD](IServiceInfo[T | TD]):
    '''
    Call `getattr()` from current `ServiceProvider`.
    '''

    __slots__ = ('_getattr_args',)
    _UNSET = object()

    def __init__(self, attr_name: str, attr_default: TD=_UNSET) -> None:
        super().__init__()
        self._getattr_args = (attr_name,) if attr_default is self._UNSET else (attr_name, attr_default)

    def __repr__(self) -> str:
        getattr_args = ', '.join(repr(x) for x in self._getattr_args)
        return f'<(ioc) => getattr(ioc, {getattr_args})>'

    @override
    def get_service(self, provider: IServiceProvider) -> T | TD:
        return getattr(provider, *self._getattr_args) # type: ignore


class ValueServiceInfo[T](IServiceInfo[T]):
    '''a `IServiceInfo` use for get fixed value.'''

    __slots__ = ('_value',)

    def __init__(self, value: T) -> None:
        self._value = value

    def __repr__(self) -> str:
        return f'<(_) => {self._value!r}>'

    @override
    def get_service(self, provider: IServiceProvider) -> T:
        return self._value


class BindedServiceInfo(IServiceInfo[object]):
    '''a `IServiceInfo` use for get value from target key.'''

    __slots__ = ('_target_key',)

    def __init__(self, target_key: Hashable) -> None:
        self._target_key = target_key

    def __repr__(self) -> str:
        return f'<(ioc) => ioc[{self._target_key!r}]>'

    @override
    def get_service(self, provider: IServiceProvider) -> object:
        return provider[self._target_key]


class FactoryServiceInfo[T](IServiceInfo[T]):
    __slots__ = ('_factory')

    def __init__(self, factory: Factory[T]) -> None:
        self._factory = factory

    @override
    def get_service(self, provider: IServiceProvider) -> T:
        return self._factory(provider)


class BoundServiceInfo[T](IServiceInfo[T]):
    __slots__ = (
        '_service_info',
        '_service_provider'
    )

    def __init__(self, service_provider: IServiceProvider, service_info: IServiceInfo[T]) -> None:
        self._service_info = service_info
        self._service_provider = service_provider

    def __repr__(self) -> str:
        return f'<Bound {self._service_info!r}>'

    @override
    def get_service(self, provider: IServiceProvider) -> T:
        return self._service_info.get_service(self._service_provider)

    @staticmethod
    def wrap(service_provider: IServiceProvider, service_info: IServiceInfo[T]) -> 'BoundServiceInfo[T]':
        if type(service_info) is BoundServiceInfo:
            service_info = service_info._service_info
        return BoundServiceInfo(service_provider, service_info)


class LifetimeServiceInfo[T](IServiceInfo[T]):
    __slots__ = (
        '_service_info', '_lifetime',
        # for not transient
        '_lock',
        # for singleton
        '_cached_value',
        # for scoped
        '_scoped_key',
    )

    _NOT_ALLOWED_KEYS = frozenset([
        Symbols.cache,
    ])

    def __init__(self, *,
            lifetime: LifeTime,
            service_provider: IServiceProvider | None,
            key: Hashable,
            service_info: IServiceInfo[T],
            scoped_key: Hashable | None = None,
        ) -> None:

        if key in self._NOT_ALLOWED_KEYS:
            raise ValueError(f'Key {key!r} is not allowed')

        if lifetime == LifeTime.singleton:
            assert service_provider is not None
            # service_provider is required when the lifetime is singleton
            service_info = BoundServiceInfo.wrap(service_provider, service_info)
            # the resolved value maybe a None, so we should cache it as a tuple.
            self._cached_value: tuple[T] | None = None

        self._lifetime = lifetime
        self._service_info = service_info
        self._scoped_key = scoped_key if scoped_key is not None else self

        if self._lifetime != LifeTime.transient:
            self._lock = RLock()
        else:
            self._lock = _NULL_CONTEXT

    def __repr__(self) -> str:
        return f'<{self._lifetime} service from {self._service_info!r}>'

    @override
    def get_service(self, provider: IServiceProvider) -> T:
        if self._lifetime is LifeTime.transient:
            return self._create(provider)

        if self._lifetime is LifeTime.scoped:
            return self._from_scoped(provider)

        if self._lifetime is LifeTime.singleton:
            return self._from_singleton(provider)

        raise NotImplementedError(f'what is {self._lifetime}?')

    def _from_scoped(self, provider: IServiceProvider) -> T:
        scoped_key = self._scoped_key
        cache = provider[Symbols.cache]
        try:
            return cache[scoped_key]
        except KeyError:
            pass
        with self._lock:
            try:
                return cache[scoped_key]
            except KeyError:
                pass
            service = self._create(provider)
            cache[scoped_key] = service
            return service

    def _from_singleton(self, provider: IServiceProvider) -> T:
        if (cached_value := self._cached_value) is None:
            with self._lock:
                if (cached_value := self._cached_value) is None:
                    self._cached_value = cached_value = (self._create(provider),)
        return cached_value[0]

    def _create(self, provider: IServiceProvider) -> T:
        '''
        return the finally service instance.
        '''
        return self._service_info.get_service(provider)


class GetOrDefaultServiceInfo[TD](IServiceInfo[object | TD]):
    _UNSET = object()
    __slots__ = ('_key', '_default')

    def __init__(self, key: Hashable, default: TD=_UNSET) -> None:
        self._key = key
        self._default = default

    def __repr__(self) -> str:
        if self.has_default():
            return f'<(ioc) => ioc.get({self._key!r}, {self._default!r})>'
        else:
            return f'<(ioc) => ioc[{self._key!r}]>'

    @override
    def get_service(self, provider: IServiceProvider) -> object | TD:
        return self.get_service_by_key(provider, self._key)

    def get_service_by_key(self, provider: IServiceProvider, key: Hashable) -> object | TD:
        if self.has_default():
            return provider.get(key, self._default)
        else:
            return provider[key]

    def has_default(self) -> bool:
        return self._default is not self._UNSET


class GetManyServiceInfo[T](IServiceInfo[list[T]]):
    '''
    Get many services from single key.
    '''
    __slots__ = ('_key',)

    def __init__(self, key: Hashable) -> None:
        self._key = key

    def __repr__(self) -> str:
        return f'<(ioc) => ioc.get_many({self._key!r})>'

    @override
    def get_service(self, provider: IServiceProvider) -> list[T]:
        return provider.get_many(self._key)


class GetGroupServiceInfo(IServiceInfo[tuple[Any, ...]]):
    __slots__ = ('_keys',)

    def __init__(self, keys: Iterable[Hashable]) -> None:
        super().__init__()
        self._keys = tuple(keys)

    @override
    def get_service(self, provider: IServiceProvider) -> tuple[Any, ...]:
        return tuple(provider[k] for k in self._keys)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
# the internal utils.
# user should not import anything from this file.
# ----------

import atexit
import inspect
import itertools
import sys
from collections.abc import Iterable, Mapping
from inspect import Parameter
from logging import getLogger
from types import MappingProxyType
from typing import Annotated, Any, Callable, NoReturn, cast, get_args, get_origin, override

from ._bases import Factory, IServiceInfo, IServiceProvider, LifeTime, SupportsContext
from ._consts import SERVICEPROVIDER_NAMING_CONVENTION
from ._internal import Disposable, ProviderOptions
from ._service_info import (
    GetGroupServiceInfo,
    GetManyServiceInfo,
    GetOrDefaultServiceInfo,
    LifetimeServiceInfo,
    ProviderServiceInfo,
    ValueServiceInfo,
)
from .annotations import InjectBy, InjectByGroup, InjectFrom, InjectWithValue
from .err import ServiceNotFoundError
from .keys import NamedType, _NamedTypeListKey
from .symbols import Symbols

_logger = getLogger(__name__)

def get_module_name(fr: inspect.FrameInfo) -> str:
    '''
    Get module name from frame info
    '''
    mo = inspect.getmodule(fr.frame)
    name = '<stdin>' if mo is None else mo.__name__
    return name

def get_frameinfos(*,
        context: int=1, exclude_anyioc_frames: bool=True
    ) -> list[inspect.FrameInfo]:
    frs = inspect.stack(context=context)[1:] # exclude get_frameinfos
    if exclude_anyioc_frames:
        frs = list(itertools.dropwhile(lambda f: get_module_name(f).partition('.')[0] == 'anyioc', frs))
    return frs

def dispose_at_exit(provider: IServiceProvider) -> Disposable:
    '''
    Register `provider.__exit__()` into `atexit` module.

    Returns a `Disposable` object to unregister and call `provider.__exit__()`.
    '''
    def callback() -> None:
        provider.__exit__(*sys.exc_info())
    def unregister() -> None:
        callback()
        atexit.unregister(callback)
    atexit.register(callback)
    return Disposable(unregister)


class NamedTypeGetOrDefaultServiceInfo(GetOrDefaultServiceInfo):
    __slots__ = ()

    def __init__(self, key: NamedType, default: object = GetOrDefaultServiceInfo._UNSET) -> None:
        super().__init__(key, default)

    @override
    def get_service(self, provider: IServiceProvider) -> object:
        key = cast(NamedType, self._key)
        try:
            return provider[key]
        except ServiceNotFoundError:
            # fallback to type only.
            return self.get_service_by_key(provider, cast(NamedType, key).type)


class FallbackToAutoCallTypeInit(NamedTypeGetOrDefaultServiceInfo):
    __slots__ = ()

    def get_service(self, provider: IServiceProvider) -> object:
        try:
            return super().get_service(provider)
        except ServiceNotFoundError:
            return wrap_signature(cast(NamedType, self._key).type, follow=True)(provider)


def get_type_and_metadatas(annotation: object) -> tuple[Any, tuple[Any, ...]]:
    assert annotation is not Parameter.empty
    if get_origin(annotation) is Annotated:
        args = get_args(annotation)
        return args[0], args[1:]
    else:
        return annotation, ()

_PARAMETER_KINDS_SINGLE_VALUE = (
    Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY)
_PARAMETER_KINDS_POSITIONAL = (
    Parameter.POSITIONAL_ONLY, Parameter.VAR_POSITIONAL)
_PARAMETER_KINDS_KEYWORD = (
    Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD)

def wrap_signature[R](func: Callable[..., R], *,
        follow: bool = False,
        override_kwargs: Mapping[str, Any] | None = None,
    ) -> Factory[R]:
    '''
    wrap the function to single argument function.

    unlike the `inject*` series of utils, this is used for implicit convert.
    '''

    if override_kwargs is None:
        override_kwargs = _EMPTY_STR_MAPPING

    sign = inspect.signature(func)
    params = list(sign.parameters.values())
    if len(params) > 1:
        params = [p for p in params if p.kind != Parameter.VAR_KEYWORD or p.annotation is not Parameter.empty]
    if len(params) > 1:
        params = [p for p in params if p.kind != Parameter.VAR_POSITIONAL]

    def get_injectinfo_from_annotation(metadatas: Iterable[Any]) \
            -> InjectBy | InjectByGroup | InjectWithValue | InjectFrom | None:
        '''
        Get Inject annotation from parameter annotation.
        '''
        if sis := [x for x in metadatas if isinstance(x, (InjectBy, InjectByGroup, InjectWithValue, InjectFrom))]:
            if len(sis) > 1:
                _logger.warning('Too many annotated InjectBy')
            return sis[0]

    def get_adapter(param: Parameter) -> ParameterAdapter | None:

        if param.kind in (Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY):
            try:
                return ParameterAdapter(param.name, ValueServiceInfo(override_kwargs[param.name]))
            except KeyError:
                pass

        if param.annotation is not Parameter.empty:
            tp, md = get_type_and_metadatas(param.annotation)
            ji = get_injectinfo_from_annotation(md)

            def raises_for_parameter() -> NoReturn:
                raise TypeError(f'{type(ji)} is not allowed on {param.kind.name} parameter')

            if param.kind is Parameter.VAR_KEYWORD:
                if ji:
                    raises_for_parameter()
                return ParameterAdapter(param.name, ValueServiceInfo(tp), unpack=True)

            elif param.kind == Parameter.VAR_POSITIONAL:
                match ji:

                    case None:
                        # create ServiceInfo for type annotation
                        si = GetManyServiceInfo(tp)

                    case InjectBy() as jb if jb.has_name():
                        raise ValueError('name is invalid for VAR_POSITIONAL parameter.')

                    case InjectBy(key, _, lifetime=lifetime) as jb:
                        if jb.has_default():
                            _logger.warning('default is invalid for VAR_POSITIONAL parameter.')
                        si = GetManyServiceInfo(key)
                        if lifetime != LifeTime.transient:
                            si = LifetimeServiceInfo(service_provider=None, key=None,
                                service_info=si,
                                lifetime=lifetime,
                                scoped_key=jb,
                            )

                    case InjectByGroup(keys):
                        si = GetGroupServiceInfo(keys)

                    case InjectWithValue() | InjectFrom():
                        raises_for_parameter()

                    case _:
                        raise NotImplementedError

                return ParameterAdapter(param.name, si, unpack=True)

            else:
                assert param.kind in _PARAMETER_KINDS_SINGLE_VALUE
                match ji:
                    case None:
                        # create ServiceInfo for type annotation
                        named_type = NamedType(param.name, tp)
                        ServiceInfoType = FallbackToAutoCallTypeInit if follow else NamedTypeGetOrDefaultServiceInfo
                        si = (
                            ServiceInfoType(named_type) if param.default is Parameter.empty
                            else ServiceInfoType(named_type, param.default)
                        )

                    case InjectBy(default=default, lifetime=lifetime) as jb:
                        if jb.has_key():
                            si = GetOrDefaultServiceInfo(jb.key, default)
                        else:
                            assert jb.has_name()
                            si = NamedTypeGetOrDefaultServiceInfo(NamedType(cast(str, jb.name), tp), default)
                        if lifetime != LifeTime.transient:
                            si = LifetimeServiceInfo(service_provider=None, key=None,
                                service_info=si,
                                lifetime=lifetime,
                                scoped_key=jb,
                            )

                    case InjectWithValue(value):
                        si = ValueServiceInfo(value)

                    case InjectByGroup(keys):
                        si = GetGroupServiceInfo(keys)

                    case InjectFrom(func=func):
                        from ._service_info.extra import TransientServiceInfo
                        si = TransientServiceInfo(func, service_provider=None, key=None)

                    case _:
                        raise NotImplementedError

                return ParameterAdapter(param.name, si, unpack=False)

        elif param.name in SERVICEPROVIDER_NAMING_CONVENTION and param.kind in _PARAMETER_KINDS_SINGLE_VALUE:
            return ParameterAdapter(param.name, ProviderServiceInfo.get_singleton_instance())

    param_adapters = [get_adapter(p) for p in params]

    if not params:
        return FactoryAdapter(func)

    # auto inject if only single unknown parameter:
    if len([True for x in param_adapters if x is None]) == 1:
        assert None in param_adapters, param_adapters
        index = param_adapters.index(None)
        param = params[index]
        if param.kind in _PARAMETER_KINDS_POSITIONAL:
            # does not need to wrap.
            param_adapters[index] = ParameterAdapter(param.name, ProviderServiceInfo.get_singleton_instance()) # not unpack

        else:
            assert param.kind in _PARAMETER_KINDS_KEYWORD
            param_name = 'provider' if param.kind == Parameter.VAR_KEYWORD else param.name
            try:
                adapter = ParameterAdapter(param_name, ValueServiceInfo(override_kwargs[param_name]))
            except KeyError:
                adapter = ParameterAdapter(param_name, ProviderServiceInfo.get_singleton_instance())
            param_adapters[index] = adapter

    if all(param_adapters):
        # all params are annotated with InjectBy(key=...)
        param_adapters = cast(list[ParameterAdapter], param_adapters)
        return FactoryAdapter(
            func,
            p_params=[
                pa for p, pa in zip(params, param_adapters, strict=True)
                if p.kind in _PARAMETER_KINDS_POSITIONAL
            ],
            k_params={
                pa.param_name: pa for p, pa in zip(params, param_adapters, strict=True)
                if p.kind not in _PARAMETER_KINDS_POSITIONAL
            },
        )

    else:
        raise TypeError('factory has too many parameters.')


class ParameterAdapter:
    __slots__ = (
        'param_name',
        'service_info',
        'unpack',
    )

    def __init__(self, name: str, service_info: IServiceInfo, *, unpack: bool = False) -> None:
        self.param_name = name
        self.service_info = service_info
        self.unpack = unpack

    def append_args(self, ioc: IServiceProvider, args: list[Any], /) -> None:
        val = self.service_info.get_service(ioc)
        if self.unpack:
            args.extend(val)
        else:
            args.append(val)

    def append_kwargs(self, ioc: IServiceProvider, kwargs: dict[str, Any], /) -> None:
        name = self.param_name
        if self.unpack:
            assert type(self.service_info) is ValueServiceInfo
            types: list[NamedType] = ioc.get_many(_NamedTypeListKey(self.service_info._value))
            for tp in types:
                if tp.name not in kwargs:
                    kwargs[tp.name] = ioc[tp]
        else:
            if name not in kwargs: # do not overwrite
                kwargs[name] = self.service_info.get_service(ioc)

_EMPTY_STR_MAPPING: Mapping[str, Any] = MappingProxyType({})

class FactoryAdapter[R](Factory[R]):
    __slots__ = (
        'func',
        'p_params',
        'k_params',
        'origin_func'
    )

    def __init__(self, func: Callable[..., R],
            p_params: Iterable[ParameterAdapter] = (),
            k_params: Mapping[str, ParameterAdapter] = _EMPTY_STR_MAPPING,
        ) -> None:
        self.func = func
        self.p_params = p_params
        self.k_params = k_params
        self.origin_func = func.func if isinstance(func, FactoryAdapter) else func

    def __call__(self, ioc: IServiceProvider, /) -> R:
        if self.p_params:
            args = []
            for param in self.p_params:
                param.append_args(ioc, args)
        else:
            args = ()

        if self.k_params:
            kwargs = {}
            for param in self.k_params.values():
                param.append_kwargs(ioc, kwargs)
        else:
            kwargs = _EMPTY_STR_MAPPING

        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return f'<Adapter of {self.origin_func!r} at {hex(id(self))}>'


def create_service[T](
        provider: IServiceProvider,
        factory: Factory[T],
        options: ProviderOptions | None = None,
    ) -> T:

    options = provider[Symbols.provider_options] if options is None else options

    service = factory(provider)
    if options['auto_enter']:
        wrapped = getattr(factory, 'origin_func', factory)
        # We must ensure that the original object is a ContextManager.
        # If the original object is a factory function and
        # the ContextManager service is merely the return value of that function,
        # then __enter__ should not be called automatically.
        if isinstance(wrapped, SupportsContext) and isinstance(service, SupportsContext):
            service = provider.enter(service)
    return service # type: ignore



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import logging
from typing import Callable

from ._utils import get_module_name as _get_module_name
from .ioc import IServiceProvider
from .symbols import Symbols


def auto_enter[R](func: Callable[..., R]) -> Callable[[IServiceProvider], R]:
    '''
    auto enter the context manager when it created.

    the signature of func should be `(ioc) => any`.
    '''
    def new_func(ioc: IServiceProvider, /) -> R:
        mgr = func(ioc)
        rv = ioc.enter(mgr) # type: ignore
        return rv

    return new_func

def get_logger(ioc: IServiceProvider, /) -> logging.Logger:
    '''
    a helper that use to get logger from ioc.

    Usage:

    ``` py
    ioc.register_transient('logger', get_logger) # use transient to ensure no cache
    logger = ioc['logger']
    assert logger.name == __name__ # the logger should have module name
    ```
    '''
    fr = ioc[Symbols.caller_frame]
    name = _get_module_name(fr)
    return logging.getLogger(name)

def is_root(provider: IServiceProvider) -> bool:
    '''
    Test is the IServiceProvider is the root provider or not.
    '''
    return provider[Symbols.provider_root] is provider

def get_scope_depth(provider: IServiceProvider) -> int:
    '''
    Get the depth of scopes.

    The root provider is 0.
    '''
    depth = 0
    root = provider[Symbols.provider_root]
    while provider is not root:
        provider = provider[Symbols.provider_parent]
        depth += 1
    return depth



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------


import inspect
from typing import TYPE_CHECKING, Any

from ._bases import IServiceProvider
from ._internal import LockedMapping, ProviderOptions
from ._primitive_symbol import TypedSymbol

if TYPE_CHECKING:
    from . import ioc_resolver  # noqa: F401


class Symbols:
    '''
    The symbols for ServiceProvider internal uses.

    All keys are predefined in `ServiceProvider`.
    overwrite those keys will break the expected behavior.
    '''

    # current scoped `IServiceProvider`
    provider = TypedSymbol[IServiceProvider]('provider')

    # the root `IServiceProvider`
    provider_root = TypedSymbol[IServiceProvider]('provider_root')

    # the parent of current `IServiceProvider`, or `None` for the root `IServiceProvider`.
    provider_parent = TypedSymbol[IServiceProvider | None]('provider_parent')

    # the cache dict to store scoped instances
    cache = TypedSymbol[LockedMapping[Any, Any]]('cache')

    # the missing resolver from `IServiceProvider`
    missing_resolver = TypedSymbol['ioc_resolver.ServiceInfoChainResolver']('missing_resolver')

    # get frame info of caller
    caller_frame = TypedSymbol[inspect.FrameInfo]('caller_frame')

    # the options for the `ServiceProvider`, value is a dict like object.
    provider_options = TypedSymbol[ProviderOptions]('provider_options')

    # is current stage of the `IServiceProvider` is initializing
    at_init = TypedSymbol[bool]('at_init')


__all__ = ['Symbols']



# -*- coding: utf-8 -*-
# 
# Copyright (c) 2025~2999 - Cologler <skyoflw@gmail.com>
# ----------
# 
# ----------

from dataclasses import dataclass, field
from typing import Callable, Hashable, Iterable

from ._bases import LifeTime
from ._service_info import GetOrDefaultServiceInfo

_UNSET_KEY = object()


@dataclass(frozen=True, slots=True, eq=False)
class InjectBy:
    '''
    Inject args by key.

    Equals:

    ```
    provider.get(key, default) if has_default() else provider[key]
    ```

    For VAR_POSITIONAL parameter, this equals:

    ```
    * provider.get_many(key)
    ```
    '''

    key: Hashable = field(default=_UNSET_KEY)
    default: object = field(default=GetOrDefaultServiceInfo._UNSET)
    # kwonly:
    lifetime: LifeTime = field(default=LifeTime.transient, kw_only=True)
    name: str | None = field(default=None, kw_only=True)

    def __post_init__(self) -> None:
        if self.lifetime == LifeTime.singleton:
            # we don't known which IServiceProvider own this.
            raise ValueError('Singleton lifetime for InjectBy is not allowed.')

        if self.has_key() and self.has_name():
            raise ValueError('key and name cannot use togeter.')

        if not self.has_key() and not self.has_name():
            raise ValueError('Missing key or name.')

    def has_key(self) -> bool:
        return self.key is not _UNSET_KEY

    def has_name(self) -> bool:
        return self.name is not None

    def has_default(self) -> bool:
        return self.default is not GetOrDefaultServiceInfo._UNSET


@dataclass(frozen=True, slots=True, eq=False)
class InjectByGroup:
    '''
    Inject args as tuple group.

    Equals:

    ```
    tuple(provider[k] for k in keys)
    ```

    For VAR_POSITIONAL parameter, this equals:

    ```
    * tuple(provider[k] for k in keys)
    ```
    '''
    keys: Iterable[Hashable]


@dataclass(frozen=True, slots=True, eq=False)
class InjectWithValue:
    '''
    Inject with the fixed value.

    Equals:

    ```
    value
    ```

    For VAR_POSITIONAL parameter, this is not allowed.
    '''
    value: object


@dataclass(frozen=True, slots=True, eq=False)
class InjectFrom:
    '''
    Inject from a callable.
    '''

    func: Callable[..., object]


__all__ = [
    'InjectBy',
    'InjectByGroup',
    'InjectFrom',
    'InjectWithValue',
]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2025~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import types
from dataclasses import dataclass
from typing import Type

from ._primitive_symbol import TypedSymbol


@dataclass(frozen=True, slots=True)
class _NamedTypeListKey:
    '''
    Internal use only.
    '''
    type: type


@dataclass(frozen=True, slots=True)
class NamedType[T]:
    '''
    The NamedType can inject into the `**kwargs`.
    '''
    name: str
    type: Type[T]

    def __post_init__(self) -> None:
        if not isinstance(self.type, (type, types.GenericAlias)):
            raise TypeError(self.type)


__all__ = [
    'TypedSymbol',
    'NamedType',
]
